<?php
/**
 * OpenTTDLib
 *  A PHP-Class to query OpenTTD Servers
 *
 * You can use this class freely for your own site.
 *  It was part of OpenTTD and therefore released under the GNU GPL v2.
 *  basicly you are free to alter it how ever you like, but
 *  you should always make your code available on request.
 *
 * Please refer to docs/ or {@link http://codecubes.org/wiki/OpenTTDLib} for help and examples
 *
 * please report bugs, problems and feature requests at {@link http://codecubes.org}
 *
 * @author	dihedral <dihedral@codecubes.org>
 * @author	TrueLight <truelight@openttd.org>
 * @link	http://codecubes.org
 * @version $Id: OpenTTDLib.php 359 2009-04-29 16:55:07Z dih $
 * @license	http://www.gnu.org/licenses/gpl.html GNU General Public License v2
 *
 * @package OpenTTDLib 
 **/

/**
 * OpenTTDLib a php class to query a running OpenTTD server
 * @author	dihedral <dihedral@codecubes.org>
 * @author	TrueLight <truelight@openttd.org>
 * @license	http://www.gnu.org/licenses/gpl.html GNU General Public License v2
 *
 * @package OpenTTDLib
 **/
class OpenTTDLib
{	
	/**#@+
	* Query Method
	* @see OpenTTDLib::queryServer()
	**/
	const QUERYTYPE_ALL					= 0;
	const QUERYTYPE_INFO				= 1;
	const QUERYTYPE_DETAIL				= 2;
	const QUERYTYPE_NEWGRF				= 3;
	/**#@-*/
	
	const RETURNTYPE_ARRAY				= 0;
	const RETURNTYPE_OBJECT				= 1;
	
	/**#@+
	 * Option constants, use with {@link OpenTTDLib::setOption()} and {@link OpenTTDLib::getOption()}
	 * @var string
	 **/
	const OPTION_CONVERT_MAPSET			= 'CONVERT_MAPSET';
	const OPTION_CONVERT_NETLANG		= 'CONVERT_NETLANG';
	const OPTION_CONVERT_VEHICLETYPE	= 'CONVERT_VEHICLETYPE';
	const OPTION_CONVERT_STATIONTYPE	= 'CONVERT_STATIONTYPE';
	
	const OPTION_NEWGRF_MAX_REQUEST		= 'NEWGRF_MAX_REQUEST';
	
	const OPTION_CACHE_NEWGRF			= 'CACHE_NEWGRF';
	const OPTION_CACHE_GAME				= 'CACHE_GAME';
	const OPTION_CACHE_FILE				= 'CACHE_FILE';
	const OPTION_CACHE_UPDATE			= 'CACHE_UPDATE';
	
	const OPTION_RETURNTYPE				= 'RETURNTYPE';
	const OPTION_QUERYTYPE				= 'QUERYTYPE';
	const OPTION_TIMEOUT				= 'TIMEOUT';
	/**#@-*/
	
	/**#@+
	 * respective default values to the options
	 **/
	const OPTION_DEFAULT_CONVERT_MAPSET			= false;
	const OPTION_DEFAULT_CONVERT_NETLANG		= false;
	const OPTION_DEFAULT_CONVERT_VEHICLETYPE	= false;
	const OPTION_DEFAULT_CONVERT_STATIONTYPE	= false;
	
	const OPTION_DEFAULT_NEWGRF_MAX_REQUEST		= 10;
	
	const OPTION_DEFAULT_CACHE_NEWGRF			= true;
	const OPTION_DEFAULT_CACHE_GAME				= false;
	const OPTION_DEFAULT_CACHE_FILE				= 'cache.xml';
	const OPTION_DEFAULT_CACHE_UPDATE			= true;
	
	const OPTION_DEFAULT_RETURNTYPE				= self::RETURNTYPE_OBJECT;
	const OPTION_DEFAULT_QUERYTYPE				= self::QUERYTYPE_ALL;
	const OPTION_DEFAULT_TIMEOUT				= 5;
	/**#@-*/
	
	/* BEGINING OF OPENTTD RELATED ENUMS */
	
	/**#@+
	* Network Language enum as found in network.h - 0.5 only up to FRENCH (0-3)
	**/
	private  $_netlang = array(
		'ANY',
		'ENGLISH',
		'GERMAN',
		'FRENCH',
		'BRAZILIAN',
		'BULGARIAN',
		'CHINESE',
		'CZECH',
		'DANISH',
		'DUTCH',
		'ESPARANTO',
		'FINNISH',
		'HUNGARIAN',
		'ICELANDIC',
		'ITALIAN',
		'JAPANESE',
		'KOREAN',
		'LITHUANIAN',
		'NORWEGIAN',
		'POLISH',
		'PORTUGUESE',
		'ROMANIAN',
		'RUSSIAN',
		'SLOVAK',
		'SLOVENIAN',
		'SPANISH',
		'SWEDISH',
		'TURKISH',
		'UKRAINIAN',
	);
	/**#@-*/
	
	/**#@+
	* Map Set enum
	**/
	private $_mapset = array(
		'TEMPARATE',
		'ARCTIC',
		'DESERT',
		'TOYLAND',
	);
	/**#@-*/
	
	/**#@+
	* Vehicle Type enum
	**/
	private $_vehicletype = array(
		'TRAIN',
		'TRUCK',
		'BUS',
		'AIRCRAFT',
		'SHIP',
	);
	/**#@-*/
	
	/**#@+
	* Station Type enum
	**/
	private $_stationtype = array(
		'TRAIN',
		'TRUCK_STOP',
		'BUS_STOP',
		'AIRPORT',
		'DOCK',
	);
	/**#@-*/
	
	/* END OF OPENTTD RELATED ENUMS */
	
	
	/**
	 * @access private
	 * @var string $_rev revision number
	 **/
	private $_rev = '$Rev: 359 $';
	
	/**
	 * @access private
	 * @var	resource $_fp filepointer as returned by fsockopen()
	 **/
	private $_fp;
	
	/**
	 * @access private
	 * @var	string	$_host	hostname or ip address of server
	 **/
	private $_host = 'localhost';
	
	/**
	 * @access private
	 * @var	integer	$_port port number the server is listening on
	 **/
	private $_port = 3979;
	
	/**
	 * @access private
	 * @var array $_option holds user set options
	 **/
	private $_option = array();
	
	/**
	 * @access private
	 * @var boolean $_connected states weather a connection is active or not
	 **/
	private $_connected;
	
	/**
	 * @access private
	 * @var array $_info holds the processed information from a OpenTTDLibPacket::UDP_SERVER_RESPONSE packet
	 **/
	protected $_info = null;
	
	/**
	 * @access private
	 * @var array $_detail holds the processed details from a OpenTTDLibPacket::UDP_SERVER_DETAIL_INFO packet
	 **/
	protected $_detail = null;
	
	/**
	 * @access private
	 * @var array $_newgrf holds the processed newgrf details from a OpenTTDLibPacket::UDP_SERVER_NEWGRFS packet
	 **/
	protected $_newgrf = array();
	
	/**
	 * @access private
	 * @var array $_newgrf_request id and md5sum for grf's still to be requested from the server
	 **/
	protected $_newgrf_request = array();
	
	/**
	 * @access private
	 * @var OpenTTDLibCache $_cache the cache object, when cacheing is enabled
	 **/
	protected $_cache = null;
	
	/**
	 * @access private
	 * @var boolean $_has_info specifies if data is on the internal buffer
	 **/
	protected $_has_info;
	
	
	/**
	 * @access private
	 * @var integer $_timeout_start_time holds the timespame when we started querying the server
	 **/
	private $_timeout_start_time = 0;
	
	/**
	 * @access private
	 * @var OpenTTDLibPacket $_packet the received packet data (@link OpenTTDLibPacket}
	 **/
	protected $_packet;

	/**
	 * prepares the class with certain variables and values
	 * @access public
	 * @return void
	 **/
	public function __construct()
	{
		$this->connected	= false;
		$this->has_info		= false;
	}
	
	/**
	 * Connect to a given IP/Port
	 * @access	public
	 * @param	string	$host	hostname or ip address of server to connect to
	 * @param	integer	$port	port the server is listening on
	 * @return	boolan			true on success
	 * @throws	OpenTTDLibException if fsockopen fails
	 **/
	public function connect( $host=null, $port=null )
	{
		if( !is_null( $host ) )
		{
			$this->_host = $host;
		}
		
		if( !is_null( $port ) )
		{
			$this->_port = $port;
		}
		
		$this->_has_info	= false;
		
		/* Create a socket, UDP mostly returns true (only not on failed host lookups) */
		if( ( $this->_fp = @fsockopen( 'udp://'. $this->_host, $this->_port, $errno, $errstr, 1 ) ) )
		{
			/* Make the socket non-blocking */
			stream_set_blocking( $this->_fp, false );
			$this->_connected = true;
		}
		else
		{
			$this->_connected = false;
			throw new OpenTTDLibException( $this->_host, $this->_port, OpenTTDLibException::ERROR_CONNECT_FAILD, $errstr );
		}
		
		return $this->_connected;
	}
	
	/**
	 * Disconnect from the server, but only if a connection is still open
	 * @access	public
	 * @return	void
	 **/
	public function disconnect()
	{
		/* only disconnect if we still are connected */
		if( $this->isConnected() )
		{
			$this->_connected = false;
			
			if ( $this->_fp )
			{
				fclose( $this->_fp );
				unset( $this->_fp );
			}
		}
	}
	
	/**
	 * initialize the cache, but only if {@link OpenTTDLib::OPTION_CACHE_NEWGRF} or {@link OpenTTDLib::OPTION_CACHE_GAME} is true
	 * @access public
	 * @param boolean $force if $this->_cache is already an instance of OpenTTDLibCache overwrite it
	 * this does not override {@link OpenTTDLib::OPTION_CACHE_NEWGRF} or {@link OpenTTDLib::OPTION_CACHE_GAME}
	 * @return void
	 **/
	public function initCache( $force=false )
	{
		if( is_null( $this->_cache ) || $force )
		{
			if( $this->getOption( self::OPTION_CACHE_NEWGRF ) || $this->getOption( self::OPTION_CACHE_GAME ) )
			{
				$this->_cache = new OpenTTDLibCache( $this->getOption( self::OPTION_CACHE_FILE ) );
			}
		}
	}
	
	/**
	 * save the cache to it's file, if options are set, and $this->_cache is an instance of OpenTTDLibCache
	 * @access public
	 * @return void
	 **/
	public function updateCache()
	{
		if( ( $this->_cache instanceof OpenTTDLibCache && $this->getOption( self::OPTION_CACHE_UPDATE ) )
			&& ( $this->getOption( self::OPTION_CACHE_NEWGRF ) || $this->getOption( self::OPTION_CACHE_NEWGRF ) ) )
		{
			$this->_cache->updateCache();
		}
	}
	
	/**
	 * checks if the option to be set is valid, if not it is ignored
	 *
	 * if $option is an array the method will itterate over the $key=>$value pairs
	 * where $key will be treated as if it were $option.
	 * the key of the array should be one of {@link OpenTTDLib::OPTION_CONVERT_MAPSET},
	 * {@link OpenTTDLib::OPTION_CONVERT_NETLANG}, {@link OpenTTDLib::OPTION_CONVERT_VEHICLETYPE},
	 * {@link OpenTTDLib::OPTION_CONVERT_STATIONTYPE}, {@link OpenTTDLib::OPTION_QUERYTYPE}, {@link OpenTTDLib::OPTION_TIMEOUT}
	 *
	 * $value is typecased to be the same type as the respective OpenTTDLib::OPTION_DEFAULT constant
	 * 
	 * @access public
	 * @param mixed $option string or array holding the separate options
	 * @param mixed $value respective value to $option or null/omitted if $option is an array()
	 * @return boolean false if any specified option was invalid
	 **/
	public function setOption( $option, $value=null )
	{
		$_return = true;
		
		if( is_array( $option ) && is_null( $value ) )
		{
			foreach( $option as $key => $val )
			{
				$_return = ( $this->setOption( $key, $val ) && $_return ) ? true : false;
			}
		}
		elseif( is_string( $option ) && !is_null( $value ) )
		{
			$_constant	= 'self::OPTION_' . $option;
			$_default	= 'self::OPTION_DEFAULT_' . $option;
			
			if( defined( $_constant ) )
			{
				settype( $value, gettype( constant( $_default ) ) );
				$this->_option[$option] = $value;
				$_return = true;
			}
			else
			{
				$_return = false;
			}
		}
		else
		{
			$_return = false;
		}
		
		return $_return;
	}
	
	/**
	 * get an option previously set with {@link OpenTTDLib::setOption()} or the respective default value
	 *
	 * check if $option is a valid option and return it's value if set, otherwise return the options default
	 *
	 * @access public
	 * @param string $option one of {@link OpenTTDLib::OPTION_CONVERT_MAPSET},
	 * {@link OpenTTDLib::OPTION_CONVERT_NETLANG}, {@link OpenTTDLib::OPTION_CONVERT_VEHICLETYPE},
	 * {@link OpenTTDLib::OPTION_CONVERT_STATIONTYPE}, {@link OpenTTDLib::OPTION_QUERYTYPE}, {@link OpenTTDLib::OPTION_TIMEOUT}
	 * @return mixed false if an invalid option was specified, the respective value otherwise
	 **/
	public function getOption( $option )
	{
		$_constant	= 'self::OPTION_' . $option;
		$_default	= 'self::OPTION_DEFAULT_' . $option;

		if( defined( $_constant ) )
		{
			if( array_key_exists( $option, $this->_option ) )
			{
				return $this->_option[$option];
			}
			else
			{
				return  constant( $_default );
			}
		}
		return false;
	}
	
	/**
	 * reset any previously set option - has to be called manually,
	 * this method is not called from within the class
	 * @access public
	 * @param mixed $option specify an option you wish to reset, or an array of options. if null reset all set options
	 * @return void
	 **/
	public function clearOption( $option=null )
	{
		if( is_null( $option ) )
		{
			$this->_option = array();
		}
		elseif( is_array( $option ) )
		{
			foreach( $option as $_opt )
			{
				$this->clearOption( $_opt );
			}
		}
		else
		{
			$_constant = 'self::OPTION_' . $option;
			if( defined( $_constant ) )
			{
				unset( $this->_option[$option] );
			}
		}
	}
	
	/**
	 * start the internal timer for checking if a timeout has occured
	 * @access public
	 * @return void
	 **/
	public function startTimer()
	{
		$this->_timeout_start_time = time();
	}
	
	/**
	 * check if we have timedout
	 * @access public
	 * @return boolean false on no timeout
	 * @throws OpenTTDLibException if we have timedout
	 **/
	public function checkTimeout()
	{
		if( time() - $this->_timeout_start_time > $this->getOption( self::OPTION_TIMEOUT ) )
		{
			throw new OpenTTDLibException( $this->_host, $this->_port, OpenTTDLibException::ERROR_CONNECT_TIMEOUT );
		}
		return false;
	}
	
	/**
	 * sends a {@link OpenTTDLibPacket::UDP_CLIENT_FIND_SERVER} packet
	 * @access private
	 * @return void
	 **/
	private function _networkSendPacketClientFindServer()
	{
		$_packet = OpenTTDLibPacket::NetworkSend_Init( OpenTTDLibPacket::UDP_CLIENT_FIND_SERVER );
		$this->sendPacket( $_packet );
	}
	
	/**
	 * sends a {@link OpenTTDLibPacket::UDP_CLIENT_DETAIL_INFO} packet
	 * @access private
	 * @return void
	 **/
	private function _networkSendPacketClientDetailInfo()
	{
		$_packet = OpenTTDLibPacket::NetworkSend_Init( OpenTTDLibPacket::UDP_CLIENT_DETAIL_INFO );
		$this->sendPacket( $_packet );
	}
	
	/**
	 * sends a {@link OpenTTDLibPacket::UDP_CLIENT_GET_NEWGRFS} packet
	 * requires to have already queried the server with {@link OpenTTDLibPacket::UDP_CLIENT_FIND_SERVER}
	 * if this has not happended beforehand, it will be called prior to continuing the grf stuff
	 * 
	 * many thanks to Rubidium for helping me getting OpenTTDLib's newgrf support working
	 * 
	 * @since version 0.3 (r45)
	 * @access private
	 * @return boolean false if there are no grfs to be queried, true if the request packet has been sent
	 **/
	private function _networkSendPacketClientGetNewGrfs()
	{
		if( is_null( $this->_info ) )
		{
			$this->_networkSendPacketClientFindServer();
			$this->waitForServerResponse();
			$this->clear();
		}
		
		if( ( $_grf_count = count( $this->_newgrf_request ) ) > 0 )
		{
			$_grf_num_request = min( $this->getOption( self::OPTION_NEWGRF_MAX_REQUEST ), $_grf_count );
			
			$_packet = OpenTTDLibPacket::NetworkSend_Init( OpenTTDLibPacket::UDP_CLIENT_GET_NEWGRFS );
			$_packet->Send_uint8( $_grf_num_request );
			
			$i = 0;
						
			foreach( $this->_newgrf_request as $_newgrf )
			{
				$_packet->Send_uint32( hexdec( $_newgrf->id ), true );
				$_md5_segs = str_split( $_newgrf->md5sum, 2 );

				foreach( $_md5_segs as $_seg )
				{
					$_packet->Send_uint8( hexdec( $_seg ) );
				}
				
				if( ++$i == $_grf_num_request )
				{
					break;
				}
			}
			
			$this->sendPacket( $_packet );
			return true;
		}
		return false;
	}
	
	/**
	 * adds a grf id and md5sum to a list of grfs to be requested from the game server
	 *
	 * @access private
	 * @param string $id grf id of the grf to be requested
	 * @param string $md5sum grf md5 checksum of the grf to be requested
	 * @return void
	 **/
	private function _addToNewGrfRequest( $id, $md5sum )
	{
		$i = count( $this->_newgrf_request );
		
		$this->_newgrf_request[$i]->id			= $id;
		$this->_newgrf_request[$i++]->md5sum	= $md5sum;
		
	}
	
	/**
	 * prepares $_grf_request for the next request, by removing already retrieved grfs from the list
	 * 
	 * many thanks to Rubidium for helping me getting OpenTTDLib's newgrf support working
	 * 
	 * @access private
	 * @return void
	 **/
	private function _updateGrfRequest()
	{
		foreach( $this->_newgrf_request as $key => $_newgrf_request )
		{
			foreach( $this->_newgrf as $_newgrf )
			{
				if( $_newgrf_request->id == $_newgrf->id && $_newgrf_request->md5sum == $_newgrf->md5sum )
				{
					unset( $this->_newgrf_request[$key] );
					break;
				}
			}
		}
		
		/* send another request with the missing grf id's and md5sums */
		if( count( $this->_newgrf_request ) > 0 )
		{
			$this->clear();
			if( $this->_networkSendPacketClientGetNewGrfs() )
			{
				$this->waitForServerResponse();
			}
		}
	}
	
	/**
	 * build the grf list of grfs to be requested, if {@link OpenTTDLib::OPTION_CACHE_NEWGRF} is true, check the cache before adding
	 *
	 * @access private
	 * @param array() $newgrfs array of objects, holding grf id and md5 checksums
	 * @return void
	 **/
	private function _buildNewGrfRequest( $newgrfs )
	{	
		foreach( $newgrfs as $newgrf )
		{
			if( $this->getOption( self::OPTION_CACHE_NEWGRF ) )
			{
				if( ( $cached_newgrf = $this->_cache->getNewGrf( $newgrf->id, $newgrf->md5sum ) ) !== false )
				{					
					$i = count( $this->_newgrf );
					
					$this->_newgrf[$i]->id		= $cached_newgrf->id;
					$this->_newgrf[$i]->md5sum	= $cached_newgrf->md5sum;
					$this->_newgrf[$i]->name	= $cached_newgrf->name;
				}
				else
				{
					$this->_addToNewGrfRequest( $newgrf->id, $newgrf->md5sum );
				}
			}
			else
			{
				$this->_addToNewGrfRequest( $newgrf->id, $newgrf->md5sum );
			}
		}
	}
	
	/**
	 * handling a packet holding ServerResponce 
	 * @access	private
	 * @return	void
	 * @throws	OpenTTDLibException if packt is not an instance of OpenTTDLibPacket
	 **/
	private function _networkRecvPacketServerResponse()
	{
		/* make sure we are working with an instance of OpenTTDLibPacket */
		$this->_checkPacket();
		
		$_info->info_version = $this->_packet->Recv_uint8();

		switch( $_info->info_version )
		{
			case 4:
				/* many thanks to Rubidium for helping me getting OpenTTDLib's newgrf support working */
				$_info->grf_count = $this->_packet->Recv_uint8();
				$_info->grf = array();
								
				for( $i = 0; $i < $_info->grf_count; $i++ )
				{					
					$_grf_id = dechex( $this->_packet->Recv_uint32( true ) );
					
					for( $j = 0; $j < 16; $j++ )
					{
						$_md5[$j] = sprintf( "%02x", $this->_packet->Recv_uint8() );
					}
					
					$_grf_md5 = implode( '', $_md5 );
					
					$_info->grf[$i]->id	= strtoupper( $_grf_id );
					$_info->grf[$i]->md5sum	= strtoupper( $_grf_md5 );
				}	
				
				/* add to grf_request array */
				$this->_buildNewGrfRequest( $_info->grf );
				
				/* Fallthrough */
				
			case 3:
				$_info->game_date	= $this->_packet->Recv_uint32();
				$_info->start_date	= $this->_packet->Recv_uint32();
				/* Fallthrough */
				
			case 2:
				$_info->companies_max	= $this->_packet->Recv_uint8();
				$_info->companies_on	= $this->_packet->Recv_uint8();
				$_info->spectators_max	= $this->_packet->Recv_uint8();
				/* Fallthrough */
				
			case 1:
				$_info->name			= $this->_packet->Recv_string();
				$_info->revision		= $this->_packet->Recv_string();
				$_info->server_lang		= $this->_packet->Recv_uint8();
				$_info->use_password	= $this->_packet->Recv_uint8();
				$_info->clients_max		= $this->_packet->Recv_uint8();
				$_info->clients_on		= $this->_packet->Recv_uint8();
				$_info->spectators_on	= $this->_packet->Recv_uint8();
				
				/*
				* 16 bit dates got scrapped in version 3 and now read there
				* currently this is brocken
				*/
				if ( $_info->info_version < 3 )
				{
					$_days_till_original_base_year = ( 365 * 1920 + 1920 / 4 - 1920 / 100 + 1920 / 400 );
					$_info->game_date	= $this->_packet->Recv_uint16() + $_days_till_original_base_year;
					$_info->start_date	= $this->_packet->Recv_uint16() + $_days_till_original_base_year;
				}
				
				$_info->map_name	= $this->_packet->Recv_string();
				$_info->map_width	= $this->_packet->Recv_uint16();
				$_info->map_height	= $this->_packet->Recv_uint16();
				$_info->map_set		= $this->_packet->Recv_uint8();
				$_info->dedicated	= $this->_packet->Recv_uint8();
	
				/* The server is online */
				$_info->online = true;
				break;
	
			default:
				/*
				* The server replied with an unsupported game info version.
				* Therefore we assume it isn't a 'correct' reply and there
				* is no working OTTD server at origin of the reply.
				*/
				$_info->online = false;
				throw new OpenTTDLibException( $this->_host, $this->_port, OpenTTDLibException::ERROR_GAME_INFO_VERSION, $info['info_version'] , $this->_packet );
				break;
		}

		/* write game info to the class */
		$this->_info = $_info;
	}

	/**
	 * handling a packet holding ServerDetailInfo
	 * @access private
	 * @return void
	 * @throws OpenTTDLibException if packt is not an instance of OpenTTDLibPacket
	 **/
	private function _networkRecvPacketServerDetailInfo()
	{
		/* make sure we are working with an instance of OpenTTDLibPacket */
		$this->_checkPacket();
		
		$num_vehicle_types = 5;
		$num_station_types = 5;
		
		/* client number counter */
		$k = 0;
		
		$_detail->detail_version = $this->_packet->Recv_uint8( );
		
		$_detail->active_players = $this->_packet->Recv_uint8( );
				
		/* make sure the structure is there, to satisfy any foreaches */
		$_detail->player     = array();
		$_detail->spectator  = array();
		$_detail->allclients = array();
		
		switch( $_detail->detail_version )
		{
			case 6:
				for( $i = 0; $i < $_detail->active_players; $i++ )
				{
					$_detail->player[$i] = $this->_detailInfoCompanyDetails();
					
					/* vehicle info */
					$_detail->player[$i]->vehicles = $this->_detailInfoCompanyVehicles($num_vehicle_types);
					
					/* station info */
					$_detail->player[$i]->stations = $this->_detailInfoCompanyStations($num_station_types);
					
					/* make sure the structure is there, to satisfy any foreaches */
					$_detail->player[$i]->client = array();
					
					$_detail->player[$i]->ai = (bool)$this->_packet->Recv_uint8();
				}
				break;
			
			case 5:
				for( $i = 0; $i < $_detail->active_players; $i++ )
				{
					$_detail->player[$i] = $this->_detailInfoCompanyDetails();
					
					/* vehicle info */
					$_detail->player[$i]->vehicles = $this->_detailInfoCompanyVehicles($num_vehicle_types);
					
					/* station info */
					$_detail->player[$i]->stations = $this->_detailInfoCompanyStations($num_station_types);
					
					/* make sure the structure is there, to satisfy any foreaches */
					$_detail->player[$i]->client = array();
				}
				break;
			
			case 4:
				for( $i = 0; $i < $_detail->active_players; $i++ )
				{
					$_detail->player[$i] = $this->_detailInfoCompanyDetails();
					
					/* vehicle info */
					$_detail->player[$i]->vehicles = $this->_detailInfoCompanyVehicles($num_vehicle_types);
					
					/* station info */
					$_detail->player[$i]->stations = $this->_detailInfoCompanyStations($num_station_types);
					
					/* make sure the structure is there, to satisfy any foreaches */
					$_detail->player[$i]->client = array();
					
					$k = 0;
					/* get a list of clients connected to this company
					 * at this point we read a boolean value from the packet, if > 0 there is another client
					 */
					while( $this->_packet->Recv_uint8() > 0 )
					{
						$_detail->player[$i]->client[$k]->client_name	= $this->_packet->Recv_string();
						$_detail->player[$i]->client[$k]->unique_id		= $this->_packet->Recv_string();
						$_detail->player[$i]->client[$k]->join_date		= $this->_packet->Recv_uint32();
						
						/* add to allclients list */
						$_detail->allclients[] =& $_detail->player[$i]->client[$k];
						
						$k++;
					}
				}
				
				/* reset the client number counter $k */
				$k = 0;
				
				/* get a list of all spectators
				 * at this point we read a boolean value from the packet, if > 0 there is another client
				 */
				while( $this->_packet->Recv_uint8() > 0 )
				{
					$_detail->spectator[$k]->client_name	= $this->_packet->Recv_string();
					$_detail->spectator[$k]->unique_id		= $this->_packet->Recv_string();
					$_detail->spectator[$k]->join_date		= $this->_packet->Recv_uint32();
					
					/* add to allclients list */
					$_detail->allclients[] =& $_detail->spectator[$k];
					$k++;
				}
				break;
		}
		
		/* write details to the class */
		$this->_detail = $_detail;
	}
	
	/**
	 * fetch company details for ServerDetailInfo
	 * @access private
	 * @return array
	 * @throws OpenTTDLibException if packt is not an instance of OpenTTDLibPacket
	 **/
	private function _detailInfoCompanyDetails()
	{
		$c = new stdClass();
		$c->current       = $this->_packet->Recv_uint8();
		$c->company_name  = $this->_packet->Recv_string();
		$c->inaugurated   = $this->_packet->Recv_uint32();
		$c->company_value = $this->_packet->Recv_uint64();
		$c->money         = $this->_packet->Recv_uint64();
		$c->income        = $this->_packet->Recv_uint64();
		$c->performance   = $this->_packet->Recv_uint16();
		$c->password      = $this->_packet->Recv_uint8();
		
		return $c;
	}
	
	/**
	 * fetch vehicle details for ServerDetailInfo
	 * @access private
	 * @return array
	 * @throws OpenTTDLibException if packt is not an instance of OpenTTDLibPacket
	 **/
	private function _detailInfoCompanyVehicles($num_vehicle_types)
	{
		for( $j = 0; $j < $num_vehicle_types; $j ++ )
		{
			$vehicles[$j] = $this->_packet->Recv_uint16();
		}
		
		return $vehicles;
	}
	
	/**
	 * fetch station details for ServerDetailInfo
	 * @access private
	 * @return array
	 * @throws OpenTTDLibException if packt is not an instance of OpenTTDLibPacket
	 **/
	private function _detailInfoCompanyStations($num_station_types)
	{
		for( $j = 0; $j < $num_station_types; $j ++ )
		{
			$stations[$j] = $this->_packet->Recv_uint16();
		}
		
		return $stations;
	}
	
	/**
	 * get the data our of a {@link OpenTTDLibPacket::UPD_SERVER_NEWGRFS} packet
	 *
	 * many thanks to Rubidium for helping me getting OpenTTDLib's newgrf support working
	 * 
	 * @access private
	 * @return void
	 * @throws OpenTTDLibException if packt is not an instance of OpenTTDLibPacket
	 **/
	private function _networkRecvPacketServerNewGrfs()
	{
		/* make sure we are working with an instance of OpenTTDLibPacket */
		$this->_checkPacket();
		
		$_new_count = $this->_packet->Recv_uint8();
		$_last_count = count( $this->_newgrf );
		
		/* for every newgrf in the packet we have received, but start counting at the total number of newgrf details we already have */
		for( $i = $_last_count; $i < $_last_count + $_new_count; $i ++ )
		{
			$_grf_id = dechex( $this->_packet->Recv_uint32( true ) );
			
			for( $j = 0; $j < 16; $j++ )
			{
				$_md5[$j] = sprintf( "%02x", $this->_packet->Recv_uint8() );
			}
			
			$_grf_md5 = implode( '', $_md5 );
			
			$_grf_name = $this->_packet->Recv_string();

			$this->_newgrf[$i]->id		= strtoupper( $_grf_id );
			$this->_newgrf[$i]->md5sum	= strtoupper( $_grf_md5 );
			$this->_newgrf[$i]->name	= $_grf_name;
			
			/* if we are using the cache - add this newgrf to the cache*/
			if( $this->getOption( self::OPTION_CACHE_NEWGRF ) )
			{
				$this->_cache->addNewGrf( $this->_newgrf[$i]->id,
										  $this->_newgrf[$i]->md5sum,
										  $this->_newgrf[$i]->name );
			}
		}
		
		$this->_updateGrfRequest();
	}
	
	/**
	 * check if the packet we have is an instance of OpenTTDLibPacket
	 * @access private
	 * @return void
	 * @throws OpenTTDLibException if packt is not an instance of OpenTTDLibPacket
	 **/
	private function _checkPacket()
	{
		/* make sure we are working with an instance of OpenTTDLibPacket */
		if( !( $this->_packet instanceof OpenTTDLibPacket ) )
		{
			throw new OpenTTDLibException( $this->_host, $this->_port, OpenTTDLibException::ERROR_PACKET_INVALID );
		}
	}
	
	/**
	 * Try to receive some data, on success a new OpenTTDLibPacket Object will be stored in $this->packet
	 * @uses	OpenTTDLibPacket to handle storeing the packet and to read data from 
	 * @access public
	 * @return	void
	 * @throws	OpenTTDLibException if {@link OpenTTDLibPacket::sizeOK()} returnes false
	 **/
	public function RecvData()
	{
		/* If no socket, no data - should never happen */
		if ( !$this->_fp )
		{
			throw new OpenTTDLibException( $this->_host, $this->_port, OpenTTDLibException::ERROR_CUSTOM, 'something went bad...' );
		}
		
		/* Check if this socket has any data on it */
		$read = array( $this->_fp );
		$changed = stream_select( $read, $write = null, $except = null, 0 );

		/* return if nothing has changed */
		if (!$changed)
		{
			return;
		}
		
		/* Read from the socket */
		$data = fread( $this->_fp, OpenTTDLibPacket::SEND_MTU );
		
		/* Set the data so it can be read */
		$packet = new OpenTTDLibPacket( $data );
		
		/* Calculate the size */
		$packet->ReadRawPacketSize();
		
		if( !$packet->sizeOK() )
		{
			$this->disconnect();
			throw new OpenTTDLibException( $this->_host, $this->_port, OpenTTDLibException::ERROR_PACKET_SIZE, '', $packet );
		}
		
		/* Tell the class that we have received some data! */
		$this->_has_info = true;
		
		/* write packet to the class */
		$this->_packet =& $packet;
	}
	
	/**
	 * process the packet data accoding to the packet type we received
	 * @see		OpenTTDLibPacket::UDP_SERVER_RESPONSE
	 * @see		OpenTTDLibPacket::UDP_SERVER_DETAIL_INFO
	 * @see		OpenTTDLibPacket::UDP_SERVER_NEWGRFS
	 * @uses	OpenTTDLibPacket
	 * @access	public
	 * @return	void
	 * @throws	OpenTTDLibException if packt is not an instance of OpenTTDLibPacket
	 * @throws	OpenTTDLibException if packt type is unknown
	 **/
	public function processPacket()
	{
		if( !( $this->_packet instanceof OpenTTDLibPacket) )
		{
			throw new OpenTTDLibException( $this->_host, $this->_port, OpenTTDLibException::ERROR_PACKET_INVALID );
		}
		
		/* Find what type it is and call the appropriate methods */
		$type = $this->_packet->Recv_uint8();

		switch( $type )
		{
			case OpenTTDLibPacket::UDP_SERVER_RESPONSE:
				$this->_networkRecvPacketServerResponse();
				break;
			
			case OpenTTDLibPacket::UDP_SERVER_DETAIL_INFO:
				$this->_networkRecvPacketServerDetailInfo();
				break;
			
			case OpenTTDLibPacket::UDP_SERVER_NEWGRFS:
				 $this->_networkRecvPacketServerNewGrfs();
				break;
			
			default:
				/* Unknown packet, drop the connection */
				$this->info['online'] = false;
				$this->disconnect();
				throw new OpenTTDLibException( $this->_host, $this->_port, OpenTTDLibException::ERROR_PACKET_TYPE, $type, $this->_packet );
				break;
		}
	}
	
	/**
	 * clears internal variables to be able to start a new query
	 * @access	public
	 * @param	boolean	$all clear ALL internal data, including received packets and resetting the connection?
	 * @return	void
	 **/
	public function clear( $all=false )
	{
		$this->_packet				= null;
		$this->_has_info			= false;
		$this->_timeout_start_time	= 0;
		
		if( $all )
		{
			$this->disconnect();
			$this->_info	= null;
			$this->_detail	= null;
			$this->_newgrf  = array();
			$this->_option	= array();
			$this->_host	= 'localhost';
			$this->_port	= 3979;
		}
	}

	/**
	 * Send a packet
	 * @access	public
	 * @param	OpenTTDLibPacket	$packet
	 * @return	integer	number of bytes sent
	 * @throws	OpenTTDLibException if sending data to host fails
	 **/
	public function sendPacket( OpenTTDLibPacket &$packet )
	{
		$packet->PrepareToSend();
		
		/* Send it */
		$r = @fwrite( $this->_fp, $packet->getData(), OpenTTDLibPacket::SEND_MTU );
		
		/* fwrite can return 0, but as we never try to send 0 bytes to the game server == is sufficient */
		if( $r == false )
		{
			throw new OpenTTDLibException( $this->_host, $this->_port, OpenTTDLibException::ERROR_PACKET_SEND, '', $packet );
		}
		
		return $r;
	}

	/**
	 * Checks if a connection is available
	 * @access	public
	 * @return	boolan	true if a connection is available
	 **/
	public function isConnected()
	{
		return $this->_connected;
	}
	
	/**
	 * get the instance of OpenTTDLibPacket which is holding the received packet data
	 * @access	public
	 * @return	mixed	OpenTTDLibPacket if $this->packet is an instance of it, null otherwise
	 **/
	public function &getPacket()
	{
		$_return = null;
		
		if( $this->_packet instanceof OpenTTDLibPacket )
		{
			$_return =& $this->_packet;
		}
		
		return $_return;
	}
	
	/**
	 * get the cache
	 * @access public
	 * @return mixed if the cache is not enabled, returnes null, else instance of OpenTTDLibCache
	 **/
	public function getCache()
	{
		return $this->_cache;
	}
	
	/**
	 * Get's the info (if any) from the internal buffer
	 * @tutorial OpenTTDLib.serverinfo.pkg
	 * @access	public
	 * @return	array 	converted info from $this->info (same structure), or an empty array
	 **/
	public function getInfo()
	{
		if( $this->getOption( self::OPTION_RETURNTYPE ) == self::RETURNTYPE_ARRAY )
		{
			return $this->_convert2array( $this->convertInfo() );
		}
		else
		{
			return $this->convertInfo();
		}
	}
	
	/**
	 * convert $this->detail and return
	 * @tutorial OpenTTDLib.serverdetail.pkg
	 * @access	public
	 * @return	array	converted $this->detail, though same structure
	 **/
	public function getDetail()
	{
		if( $this->getOption( self::OPTION_RETURNTYPE ) == self::RETURNTYPE_ARRAY )
		{
			return $this->_convert2array( $this->convertDetail() );
		}
		else
		{
			return $this->convertDetail();
		}
	}
	
	/**
	 * get the new grf details
	 * @tutorial OpenTTDLib.servernewgrf.pkg
	 * @access	public
	 * @return	array	array of returned (or cached) newgrf data (array of objects as default behaviour)
	 **/
	public function getNewGrf( $sort=false )
	{
		$_newgrf = array();
		
		if( count( $this->_newgrf ) > 0 )
		{
			$_newgrf = $this->_newgrf;

			if( $sort )
			{
				$new_g = array();
				foreach( $_newgrf as $_g )
				{
					$new_g[] = strtolower( $_g->name );
				}
				
				array_multisort( $new_g, $_newgrf );
			}
		}

		if( $this->getOption( self::OPTION_RETURNTYPE ) == self::RETURNTYPE_ARRAY )
		{
			$_newgrf = $this->_convert2array( $_newgrf );
		}
		
		return $_newgrf;
	}
	
	/**
	 * ittereates through an object and converts it to an array()
	 * @access private
	 * @param mixed $object array or object to be converted (array of objects, for example)
	 * @return array() converted $object
	 **/
	private function _convert2array( $object )
	{
		$array = array();
		foreach( $object as $key => $value )
		{
			if( is_object( $value ) || is_array( $value ) )
			{
				$array[$key] = $this->_convert2array( $value );
			}
			else
			{
				$array[$key] = $value;
			}
		}

		return (array) $array;
	}
	
	/**
	 * check there is info on the internal buffer
	 * @access	public
	 * @return	boolean true if there is info on the interneal buffer
	 **/
	public function hasInfo()
	{
		return $this->_has_info;
	}
	
	/**
	 * calculate if $year is a leapyear
	 * @access public
	 * @param	integer	$year	year to check
	 * @return	boolean			true if $year is a leapyear, false otherwise
	 **/
	public function isLeapYear( $year )
	{
		return $year % 4 == 0 && ( $year % 100 != 0 || $year % 400 == 0 );
	}
	
	/**
	 * Converts OpenTTD Date to YMD (tnx to WizKid for converting)
	 * @access public
	 * @param	integer	$date	openttd style date
	 * @return	array			date in the following format: [0] = year, [1] = month [2] = day
	 **/
	public function Date2YMD( $date )
	{
		
		/* Converted from an ingame table by WizKid */
		$OTTD_month_date_from_year_day = array (
			/* January */
			array(0,1), array(0,2), array(0,3), array(0,4), array(0,5), array(0,6), array(0,7), array(0,8), array(0,9),
			array(0,10), array(0,11), array(0,12), array(0,13), array(0,14), array(0,15), array(0,16), array(0,17),
			array(0,18), array(0,19), array(0,20), array(0,21), array(0,22), array(0,23), array(0,24), array(0,25),
			array(0,26), array(0,27), array(0,28), array(0,29), array(0,30), array(0,31),
			
			/* February */
			array(1,1), array(1,2), array(1,3), array(1,4), array(1,5), array(1,6), array(1,7), array(1,8), array(1,9),
			array(1,10), array(1,11), array(1,12), array(1,13), array(1,14), array(1,15), array(1,16), array(1,17),
			array(1,18), array(1,19), array(1,20), array(1,21), array(1,22), array(1,23), array(1,24), array(1,25),
			array(1,26), array(1,27), array(1,28), array(1,29),
			
			/* March */
			array(2,1), array(2,2), array(2,3), array(2,4), array(2,5), array(2,6), array(2,7), array(2,8), array(2,9),
			array(2,10), array(2,11), array(2,12), array(2,13), array(2,14), array(2,15), array(2,16), array(2,17),
			array(2,18), array(2,19), array(2,20), array(2,21), array(2,22), array(2,23), array(2,24), array(2,25),
			array(2,26), array(2,27), array(2,28), array(2,29), array(2,30), array(2,31),
			
			/* April */
			array(3,1), array(3,2), array(3,3), array(3,4), array(3,5), array(3,6), array(3,7), array(3,8), array(3,9),
			array(3,10), array(3,11), array(3,12), array(3,13), array(3,14), array(3,15), array(3,16), array(3,17),
			array(3,18), array(3,19), array(3,20), array(3,21), array(3,22), array(3,23), array(3,24), array(3,25),
			array(3,26), array(3,27), array(3,28), array(3,29), array(3,30),
			
			/* May */
			array(4,1), array(4,2), array(4,3), array(4,4), array(4,5), array(4,6), array(4,7), array(4,8), array(4,9),
			array(4,10), array(4,11), array(4,12), array(4,13), array(4,14), array(4,15), array(4,16), array(4,17),
			array(4,18), array(4,19), array(4,20), array(4,21), array(4,22), array(4,23), array(4,24), array(4,25),
			array(4,26), array(4,27), array(4,28), array(4,29), array(4,30), array(4,31),
			
			/* June */
			array(5,1), array(5,2), array(5,3), array(5,4), array(5,5), array(5,6), array(5,7), array(5,8), array(5,9),
			array(5,10), array(5,11), array(5,12), array(5,13), array(5,14), array(5,15), array(5,16), array(5,17),
			array(5,18), array(5,19), array(5,20), array(5,21), array(5,22), array(5,23), array(5,24), array(5,25),
			array(5,26), array(5,27), array(5,28), array(5,29), array(5,30),
			
			/* July */
			array(6,1), array(6,2), array(6,3), array(6,4), array(6,5), array(6,6), array(6,7), array(6,8), array(6,9),
			array(6,10), array(6,11), array(6,12), array(6,13), array(6,14), array(6,15), array(6,16), array(6,17),
			array(6,18), array(6,19), array(6,20), array(6,21), array(6,22), array(6,23), array(6,24), array(6,25),
			array(6,26), array(6,27), array(6,28), array(6,29), array(6,30), array(6,31),
			
			/* August */
			array(7,1), array(7,2), array(7,3), array(7,4), array(7,5), array(7,6), array(7,7), array(7,8), array(7,9),
			array(7,10), array(7,11), array(7,12), array(7,13), array(7,14), array(7,15), array(7,16), array(7,17),
			array(7,18), array(7,19), array(7,20), array(7,21), array(7,22), array(7,23), array(7,24), array(7,25),
			array(7,26), array(7,27), array(7,28), array(7,29), array(7,30), array(7,31),
			
			/* September */
			array(8,1), array(8,2), array(8,3), array(8,4), array(8,5), array(8,6), array(8,7), array(8,8), array(8,9),
			array(8,10), array(8,11), array(8,12), array(8,13), array(8,14), array(8,15), array(8,16), array(8,17),
			array(8,18), array(8,19), array(8,20), array(8,21), array(8,22), array(8,23), array(8,24), array(8,25),
			array(8,26), array(8,27), array(8,28), array(8,29), array(8,30),
			
			/* October */
			array(9,1), array(9,2), array(9,3), array(9,4), array(9,5), array(9,6), array(9,7), array(9,8), array(9,9),
			array(9,10), array(9,11), array(9,12), array(9,13), array(9,14), array(9,15), array(9,16), array(9,17),
			array(9,18), array(9,19), array(9,20), array(9,21), array(9,22), array(9,23), array(9,24), array(9,25),
			array(9,26), array(9,27), array(9,28), array(9,29), array(9,30), array(9,31),
			
			/* November */
			array(10,1), array(10,2), array(10,3), array(10,4), array(10,5), array(10,6), array(10,7), array(10,8),
			array(10,9), array(10,10), array(10,11), array(10,12), array(10,13), array(10,14), array(10,15),
			array(10,16), array(10,17), array(10,18), array(10,19), array(10,20), array(10,21), array(10,22),
			array(10,23), array(10,24), array(10,25), array(10,26), array(10,27), array(10,28), array(10,29),
			array(10,30),
			
			/* December */
			array(11,1), array(11,2), array(11,3), array(11,4), array(11,5), array(11,6), array(11,7), array(11,8),
			array(11,9), array(11,10), array(11,11), array(11,12), array(11,13), array(11,14), array(11,15),
			array(11,16), array(11,17), array(11,18), array(11,19), array(11,20), array(11,21), array(11,22),
			array(11,23), array(11,24), array(11,25), array(11,26), array(11,27), array(11,28), array(11,29),
			array(11,30), array(11,31)
		);
		
		/* Year determination in multiple steps for leap years */
		
		/* There are 97 leap years in 400 years */
		$year = 400 * floor( $date / ( 365 * 400 + 97 ) );
		$rem  = $date % ( 365 * 400 + 97 );

		/* There are 24 leap years in 100 years */
		$year += 100 * floor( $rem / (365 * 100 + 24 ) );
		$rem   = $rem % ( 365 * 100 + 24 );

		/* There is 1 leap year every 4 years */
		$year += 4 * floor( $rem / ( 365 * 4 + 1 ) );
		$rem   = $rem % ( 365 * 4 + 1 );

		while( $rem >= ( $this->isLeapYear( $year ) ? 366 : 365 ) )
		{
			$rem -= $this->isLeapYear( $year ) ? 366 : 365;
			$year++;
		}

		if ( !$this->isLeapYear( $year ) && $rem >= 31 + 28 ) $rem++;

		list ( $month, $day ) = $OTTD_month_date_from_year_day[$rem];

		return array( $year, $month + 1, $day );
	}
	
	/**
	 * turn openttd dates into a readable YYYY-MM-DD format
	 * @access private
	 * @param	integer	$value	openttd style date
	 * @return	string			YYYY-MM-DD formated date
	 **/
	private function convert_makePrettyDate( $value )
	{
		$date = $this->Date2YMD( $value );
		return sprintf( "%d-%02d-%02d", $date[0], $date[1], $date[2] );
	}
	
	/**
	 * returns the corresponding numeric value or string according to the enum $type and the requested value $value
	 * @access private
	 * @param string $type can be on of _netlang, _mapset, _vehicletype, _stationtype
	 * @param mixed $value depends on what should be returned, see return
	 * @return mixed integer if $value is string, string if $value is integer, array() if $value is null
	 **/
	private function _convert_enums( $type, $value=null )
	{
		switch( $type )
		{
			case '_netlang':
			case '_mapseet':
			case '_vehicletype':
			case '_stationtype':
				$enum = $type;
				break;
			
			default:
				return $value;
				break;
		}
		
		if( is_null( $value ) )
		{
			$r = $this->$enum;
		}
		elseif( array_key_exists( $value, $this->$enum ) )
		{
			$r = $this->{$enum}[$value];
		}
		elseif( ( $k = array_search( strtoupper( $value ), $this->$enum ) ) !== false )
		{
			$r = $k;
		}
		else
		{
			return $value;
		}
	
		return $r;
	}
	
	/**
	 * converts the numbers to the appropriate server language names and vise versa
	 * 
	 * @access	public
	 * @param	mixed $value number or string to be converted
	 * @return	mixed integer if $value is string, string if $value is integer, array() if $value is null
	 **/
	public function convertNetLang( $value=null )
	{
		return $this->_convert_enums( '_netlang', $value );
	}
	
	/**
	 * converts the numbers to the appropriate mapset names and vise versa
	 * 
	 * @access	public
	 * @param	mixed $value number or string to be converted
	 * @return	mixed integer if $value is string, string if $value is integer, array() if $value is null
	 **/
	public function convertMapSet( $value=null )
	{
		return $this->_convert_enums( '_mapset', $value );
	}
	
	/**
	 * converts the numbers to the appropriate vehicle type names and vise versa
	 * 
	 * @access	public
	 * @param	mixed $value number or string to be converted
	 * @return	mixed integer if $value is string, string if $value is integer, array() if $value is null
	 **/
	public function convertVehicleType( $value=null )
	{
		return $this->_convert_enums( '_vehicletype', $value );
	}
	
	/**
	 * converts the numbers to the appropriate station type names and vise versa
	 * 
	 * @access	public
	 * @param	mixed $value number or string to be converted
	 * @return	mixed integer if $value is string, string if $value is integer, array() if $value is null
	 **/
	public function convertStationType( $value=null )
	{
		return $this->_convert_enums( '_stationtype', $value );
	}
	
	/**
	 * converts $value according to it's internal convert table
	 * @access	private
	 * @param	string	$name	name must match the key of the internal convert table
	 * @param	mixed	$value	
	 * @return	mixed			modified $value, unmodified if name does not match a key
	 **/
	private function convert( $name, $value )
	{
		$convert_table = array (
			"game_date"		=> "DATE",
			"start_date"	=> "DATE",
			"server_lang"	=> "NETLANG",
			"map_set"		=> "MAPSET",
			"join_date"		=> "DATE",
		);
		
		if( array_key_exists( $name, $convert_table ) )
		{
			switch( $convert_table[$name] )
			{
				case "DATE":
					if( $value > 0 )
					{
						$value = $this->convert_makePrettyDate( $value );
					}
					break;
				
				case "NETLANG":
					// check the config here and not when we already are in the convert_... method
					if( $this->getOption( self::OPTION_CONVERT_NETLANG ) )
					{
						$value = $this->convertNetLang( $value );
					}
					break;
				
				case "MAPSET":
					// check the config here and not when we already are in the convert_... method
					if( $this->getOption( self::OPTION_CONVERT_MAPSET ) )
					{
						$value = $this->convertMapSet( $value );
					}
					break;
				
				default:
					$value = "Not supported";
					break;
			}
		}
		
		return $value;
	}
	
	/**
	 * passes data from $this->info to $this->convert
	 * @access	private
	 * @return	array	the processed data, sructure is the same as $this->info, only the values changed
	 **/
	private function convertInfo()
	{
		
		$info = $this->_info;
		foreach ( $info as $name => $value )
		{
			$info->$name = $this->convert( $name, $value );
		}
		
		return $info;	
	}
	
	/**
	 * passes data from $this->detail to $this->convert()
	 * @access	private
	 * @return	array	the processed data, struckture is the same as $this->detail, only the values changed
	 **/
	private function convertDetail()
	{
		$detail = $this->_detail;
		
		/* detail->player */
		for( $i = 0; $i < $detail->active_players; $i++ )
		{
			foreach( $detail->player[$i] as $key => $value )
			{
				if( is_array( $value ) )
				{
					/* detail->player ->vehicle, ->station */
					foreach( $value as $k2 => $v2 )
					{
						switch( $key )
						{
							case 'vehicles':
								if( $this->getOption( self::OPTION_CONVERT_VEHICLETYPE ) )
								{
									$_key = $this->convertVehicleType( $k2 );
									unset( $detail->player[$i]->{$key}[$k2] );
									$detail->player[$i]->{$key}[$_key] = $v2;
								}
								break;
							
							case 'stations':
								if( $this->getOption( self::OPTION_CONVERT_STATIONTYPE ) )
								{
									$_key = $this->convertStationType( $k2 );
									unset( $detail->player[$i]->{$key}[$k2] );
									$detail->player[$i]->{$key}[$_key] = $v2;
								}
								break;
							
							default:
								// do nothing for detail->player->client we do that later
								$detail->player[$i]->{$key}[$k2] = $this->convert( $k2, $v2 );
								break;
						}
					}
				}
				else
				{
					$detail->player[$i]->$key = $this->convert( $key, $value );
				}
			}
		}
		
		/* detail->allclients */
		for( $i = 0; $i < count( $detail->allclients ); $i++ )
		{
			foreach( $detail->allclients[$i] as $key => $value )
			{
				$detail->allclients[$i]->$key = $this->convert( $key, $value );
			}
		}
		
		return $detail;
	}

	/**
	 * Checks if a socket has anything that we should receive
	 * also handles the timeout and makes sure not to ittereate
	 * once an error occured or the connection was dropped
	 * @access	private
	 * @return	boolean	if true, we have reveiced data, false otherwise
	 * @throws	OpenTTDLibException if server takes too long to respond
	 **/
	private function waitForServerResponse()
	{
		$this->startTimer();
		
		while( !$this->hasInfo() && $this->isConnected() )
		{
			/* sleep a little until checking for data again */
			usleep( 100000 );
			
			/* Check if we have any data */
			$this->RecvData();
			$this->checkTimeout();
		}
		
		if( $this->hasInfo() )
		{
			$this->processPacket();
		}
		
		return $this->hasInfo();
	}
	
	/**
	 * provides a simple way to query a server, this method will handle the rest for you
	 * after this method returns, you either can fetch processed response packets from the server
	 * or an error message
	 * @tutorial	OpenTTDLib.howto.pkg
	 * @access	public
	 * @param	string	$host		hostname or ip address of the server
	 * @param	integer	$port		port number the server is listening on
	 * @param	array	$options	an array of option => value pairs, see {@link OpenTTDLib::setOption()}
	 * @return	boolean				false if an error was encountered
	 * @throws	OpenTTDLibException if query type unknown - Note: called methods also throw this exception on any error
	 **/
	public function queryServer( $host=null, $port=null, $options=array() )
	{
		/* remove all data from possible previous runs */
		$this->clear( true );
		
		/* set the options for this query */
		$this->setOption( $options );
		
		/* init the cache if we are making use of it at all */
		$this->initCache();
		
		/* default fallthrough is false */
		$fallthrough = false;
		
		if( $this->connect( $host, $port ) )
		{
			switch( $this->getOption( self::OPTION_QUERYTYPE ) )
			{
				case self::QUERYTYPE_ALL:				
					$fallthrough = true;
					/* Fallthrough */
					
				case self::QUERYTYPE_INFO:					
					$this->clear();
					$this->_networkSendPacketClientFindServer();
					$this->waitForServerResponse();
					if( !$fallthrough )
					{
						break;
					}
					//else
						/* Fallthrough */
					
				case self::QUERYTYPE_DETAIL:					
					$this->clear();
					$this->_networkSendPacketClientDetailInfo();
					$this->waitForServerResponse();
					if( !$fallthrough )
					{
						break;
					}
					//else
						/* Fallthrough */
					
				case self::QUERYTYPE_NEWGRF:
					/* many thanks to Rubidium for helping me getting OpenTTDLib's newgrf support working */
					$this->clear();
					if( $this->_networkSendPacketClientGetNewGrfs() )
					{
						$this->waitForServerResponse();
					}
					break;
				
				default:
					throw new OpenTTDLibException( $this->_host, $this->_port, OpenTTDLibpException::ERROR_CUSTOM, 'unknown query type, dont know what to send and how to receive' );
					break;
			}
		}
		
		$this->disconnect();
		
		/* save the cache to the file if we are using it at all */
		$this->updateCache();
		
		return true;
	}
	
}

?>
