<?php
/**
 * OpenTTDLib
 *  A PHP-Class to query OpenTTD Servers
 *
 * You can use this class freely for your own site.
 *  It was part of OpenTTD and therefore released under the GNU GPL v2.
 *  basicly you are free to alter it how ever you like, but
 *  you should always make your code available on request.
 *
 * Please refer to docs/ or {@link http://codecubes.org/wiki/OpenTTDLib} for help and examples
 *
 * please report bugs, problems and feature requests at {@link http://codecubes.org}
 *
 * @author	dihedral <dihedral@codecubes.org>
 * @link	http://codecubes.org
 * @version $Id: OpenTTDLibCache.php 273 2009-04-01 10:41:19Z dih $
 * @license	http://www.gnu.org/licenses/gpl.html GNU General Public License v2
 *
 * @package OpenTTDLib 
 **/

/**
 * OpenTTDLibCache handles cached newgrf details, once a grfid and grf_md5sum pair is known,
 * we dont have to query the game for it again.
 * 
 * @author	dihedral <dihedral@codecubes.org>
 * @license	http://www.gnu.org/licenses/gpl.html GNU General Public License v2
 * 
 * @since version 0.3 (r46)
 * @package OpenTTDLib
 **/
class OpenTTDLibCache
{
	
	/**
	 * @access private
	 * @var SimpleXMLElement $_cache holds the parsed xml file
	 **/
	protected $_cache;
	
	/**
	 * @access private
	 * @var string $_file path or url of the xml file
	 **/
	protected $_file;
	
	/**
	 * @access private
	 * @var boolean $_updated, set to true if anything in $_cache has been changed, this way we know if the file needs updating
	 **/
	protected $_updated = false;
	
	/**
	 * constructor of OpenTTDLIbCache
	 * 
	 * @access public
	 * @param string $file path or url of the cache file
	 * @return void
	 **/
	public function __construct( $file )
	{
		$this->_file = $file;
		$this->_initCache();
	}
	
	/**
	 * loads the file, checks and parses it's content. if the file holds no (usable) xml data, a pre defined xml string is used
	 *
	 * @access private
	 * @return void
	 **/
	private function _initCache()
	{
		if( @touch( $this->_file ) === false || !is_readable( $this->_file ) )
		{
			throw new OpenTTDLibException( '', '', OpenTTDLibException::ERROR_CACHE_NOT_READABLE, $this->_file );
		}
		
		$_xml = file_get_contents( $this->_file );
		
		try
		{
			$this->_cache = new SimpleXMLElement( $_xml );
			$children = (array) $this->_cache->children();
			if( !( array_key_exists( 'games', $children ) && array_key_exists( 'newgrfs', $children ) ) )
			{
				throw new Exception();
			}
		}
		catch( Exception $e )
		{
			$this->_cache = new SimpleXMLElement( $this->_initXML() );
		}
	}
	
	/**
	 * return a xml string that we can work with
	 * 
	 * @access private
	 * @return string
	 **/
	private function _initXML()
	{
		$_xml = '<?xml version="1.0" standalone="yes"?><cache><games/><newgrfs/></cache>';
		
		$this->_updated = true;
		
		return $_xml;
	}
	
	/**
	 * convert the xml data into pretty indented xml
	 * 
	 * @access public
	 * @param integer $indent how often $str_indent is used for a single indention
	 * @param string $str_indent char for used when indenting (i.e. space or tap)
	 * @param string $lf linefeed (i.e. *nix uses \n, win uses \r\n )
	 * @return string pretty xml data
	 **/
	public function asPrettyXML( $indent=1, $str_indent="\t", $lf="\n" )
    {
        $string = $this->_cache->asXML();
       
		/* put each element on it's own line */
        $string = preg_replace( "/>\s*</", ">$lf<", $string );

        /* each element to own array */
        $xmlArray = explode( $lf, $string );

        /* holds indentation */
        $currIndent = 0;

        /* set xml element first by shifting of initial element */
        $string = array_shift( $xmlArray ) . "\n";

        foreach( $xmlArray as $element )
		{
            
            if( preg_match( '/^<([\w])+[^>\/]*>$/U', $element ) )
			{
				/* find open only tags... add name to stack, and print to string, increment currIndent */
                $string .=  str_repeat($str_indent, $currIndent) . $element . $lf;
                $currIndent += $indent;
            }
            elseif( preg_match( '/^<\/.+>$/', $element ) )
			{
				/* find standalone closures, decrement currindent, print to string */
                $currIndent -= $indent;
                $string .=  str_repeat($str_indent, $currIndent) . $element . $lf;
            }
            else
			{
				/* find open/closed tags on the same line print to string */
                $string .=  str_repeat($str_indent, $currIndent) . $element . $lf;
            }
        }
		
		return $string;
    }
	
	/**
	 * writes data from {@link OpenTTDLibCache::asPrettyXML()} back to the cache file, but only if the cache is different from that in the file
	 *
	 * @access public
	 * @return mixed, value of file_put_contents if the xml data has changed, or true
	 **/
	public function updateCache()
	{
		if( !is_writeable( $this->_file ) )
		{
			throw new OpenTTDLibException( '', '', OpenTTDLibException::ERROR_CACHE_NOT_WRITEABLE, $this->_file );
		}
		
		return ( $this->_updated ) ? file_put_contents( $this->_file, $this->asPrettyXML() ) : true;
	}
	
	/**
	 * remove double entries, determind by grf_id and grf_md5sum
	 * @access public
	 * @return void
	 **/
	public function uniqueNewGrf()
	{		
		for( $i = 0; $i < count( $this->_cache->newgrfs->newgrf ); $i++ )
		{
			if( $this->findNewGrf( $this->_cache->newgrfs->newgrf[$i]->id, $this->_cache->newgrfs->newgrf[$i]->md5sum, $i ) !== false )
			{
				unset( $this->_cache->newgrfs->newgrf[$i] );
			}
		}
	}
	
	/**
	 * find a cached newgrf by it's id and md5sum
	 * @access public
	 * @param string $grf_id id of the newgrf to find
	 * @param string $grf_md5 md5sum of the newgrf to find
	 * @param integer $pos used by {@link OpenTTDLibNewGrfCache::unique()} to determin the internal position of the grf,
	 * there will never be a reason for you to use it.
	 * @return mixed, false when none was found, else integer of the internal position of the grf
	 **/
	public function findNewGrf( $grf_id, $grf_md5, $pos=null )
	{
		for( $i = 0; $i < count( $this->_cache->newgrfs->newgrf ); $i++ )
		{
			if( trim( strtoupper( $grf_id ) ) == trim( strtoupper( $this->_cache->newgrfs->newgrf[$i]->id ) )
				&& trim( strtoupper( $grf_md5 ) ) == trim( strtoupper( $this->_cache->newgrfs->newgrf[$i]->md5sum ) )
				&& ( is_null( $pos ) || $pos != $i )
				)
			{
				return $i;
			}
		}
		
		return false;
	}
	
	/**
	 * get newgrf data from the cache if it can be found
	 *
	 * @access public
	 * @param string $grf_id the grfid of the newgrf file
	 * @param string $grf_md5 the md5 checksum of the newgrf file
	 * @return mixed object holding the details if newgrf could be found in the cache, false otherwise
	 **/
	public function getNewGrf( $grf_id, $grf_md5 )
	{
		if( ( $pos = $this->findNewGrf( $grf_id, $grf_md5 ) ) !== false )
		{
			foreach( $this->_cache->newgrfs->newgrf[$pos] as $atrib => $value )
			{
				$_tmp->$atrib = (string) $value;
			}
			
			return $_tmp;
		}
		else
		{
			return false;
		}
	}
	
	/**
	 * add a grf to the cache
	 * @access public
	 * @param string $grf_id
	 * @param string $grf_md5
	 * @param string $name
	 * return void
	 **/
	public function addNewGrf( $grf_id, $grf_md5, $name )
	{
		$_newgrf = $this->_cache->newgrfs->addChild( 'newgrf' );
		$_newgrf->addChild( 'id', trim( strtoupper( $grf_id ) ) );
		$_newgrf->addChild( 'md5sum', trim( strtoupper( $grf_md5 ) ) );
		$_newgrf->addChild( 'name', trim( $name ) );
		$_newgrf->addChild( 'timestamp', trim( time() ) );
		
		$this->_updated = true;
	}
	
}

?>
