<?php
/**
 * OpenTTDLib
 *  A PHP-Class to query OpenTTD Servers
 *
 * You can use this class freely for your own site.
 *  All parts of OpenTTDLib are released under the GNU GPL v2
 *  basicly you are free to alter it how ever you like, but
 *  you should always make your code available on request.
 *
 * Please refer to docs/ or {@link http://codecubes.org/wiki/OpenTTDLib} for help and examples
 *
 * please report bugs, problems and feature requests at {@link http://codecubes.org}
 *
 * @author	dihedral <dihedral@codecubes.org>
 * @link	http://codecubes.org
 * @version $Id: OpenTTDLibException.php 273 2009-04-01 10:41:19Z dih $
 * @license	http://www.gnu.org/licenses/gpl.html GNU General Public License v2
 *
 * @package OpenTTDLib 
 **/

/**
 * OpenTTDLibException is as of rev24 thrown by OpenTTDLib
 * @author	dihedral <dihedral@codecubes.org>
 * @version $Id: OpenTTDLibException.php 273 2009-04-01 10:41:19Z dih $
 * @license	http://www.gnu.org/licenses/gpl.html GNU General Public License v2
 *
 * @package OpenTTDLib
 * @since rev24
 **/
class OpenTTDLibException extends Exception
{
    /**#@+
	 * $var integer error value to pass to {@link OpenTTDLibException::__construct()}
	 **/
    const ERROR_CUSTOM				= 0;
	
    const ERROR_PACKET_SIZE			= 10;
    const ERROR_PACKET_TYPE			= 11;
	const ERROR_PACKET_SEND			= 12;
	const ERROR_PACKET_INVALID		= 13;
	
    const ERROR_CONNECT_FAILD		= 20;
    const ERROR_CONNECT_TIMEOUT		= 21;
	
	const ERROR_GAME_INFO_VERSION	= 30;
	
	const ERROR_CACHE_NOT_READABLE	= 40;
	const ERROR_CACHE_NOT_WRITEABLE	= 41;
    /**#@-*/
	
	/**
	 * @var string $_rev revision number
	 **/
	private $_rev = '$Rev: 273 $';
	
	/**
	 * @var mixed OpenTTDLibPacket if available at exception, otherwise null
	 **/
    protected $_packet;
	
	/**
	 * @var string host name or ip as passed to {@link OpenTTDLib::queryServer()}
	 **/
	protected $_host;
	
	/**
	 * @var integer port number as passed to {@link OpenTTDLib::queryServer()}
	 **/
	protected $_port;
    
	/**
	 * cunstructor to OpenTTDLibException
	 * @param string	$host hostname or ip as passed to {@link OpenTTDLib::queryServer()}
	 * @param integer	$port port number as passed to {@link OpenTTDLib::queryServer()}
	 * @param integer	$code error code, use one of the {@link OpenTTDLibException::ERROR_CUSTOM ERROR constants}
	 * @param string	$message depending on $code, it's appended to the error message or dropped
	 * @param OpenTTDLibPacket $packet copy of OpenTTDLibPacket if available when Exception is thrown
	 * @access public
	 * @return void
	 **/
    public function __construct ( $host, $port, $code=null, $message="", OpenTTDLibPacket $packet=null )
	{
		if( is_null( $code ) ){
			$code = self::ERROR_CUSTOM;
		}
		
		parent::__construct( $message, $code );
		
		$this->_prependCodeMessage();
		
		$this->_host	= $host;
		$this->_port	= $port;
        $this->_packet	= $packet;
    }
    
	/**
	 * set the error message according to the error code
	 * @access private
	 * @return void
	 **/
	protected function _prependCodeMessage()
	{
		switch( $this->code ){
			case self::ERROR_PACKET_SIZE:
				$this->message = "bad packet size";
				break;
			
			case self::ERROR_PACKET_TYPE:
				$this->message = "unknown packet type: " . $this->message;
				break;
			
			case self::ERROR_PACKET_SEND:
				$this->message = "sending packet failed";
				break;
			
			case self::ERROR_PACKET_INVALID:
				$this->message = "packet not an instance of OpenTTDLibPacket";
				break;
			
			case self::ERROR_CONNECT_FAILD:
				$this->message = "unable to establish connection \n" . $this->message;
				break;
			
			case self::ERROR_CONNECT_TIMEOUT:
				$this->message = "connection timed out";
				break;
			
			case self::ERROR_GAME_INFO_VERSION:
				$this->message = "unsupported game info version: " . $this->message;
				break;
			
			case self::ERROR_CACHE_NOT_READABLE:
				$this->message = "cannot read from: " . $this->message;
				break;
			
			case self::ERROR_CACHE_NOT_WRITEABLE:
				$this->message = "cannot write to: " . $this->message;
				break;
			
			default:
				$this->message = "custom message: \n" . $this->message;
				break;
		}
	}
	
	/**
	 * logic of how the class in converted to a string
	 * @access public
	 * @return string the most relevant data in the Exception
	 **/
	public function __toString()
	{
		return "{$this->file}:{$this->line} Error#{$this->code} Message:{$this->message}\n\nTrace:\n" . $this->getTraceAsString();
	}
    
	/**
	 * returns all the instances data base64 encoded, making it easy find out what went wrong
	 * when provided along with a bug report
	 * @access public
	 * @return string
	 **/
	public function dumpForDebug()
	{
		return chunk_split( base64_encode( serialize( $this ) ) );
	}

}

?>