<?php
/**
 * OpenTTDLibPacket
 *  A PHP-Class to to handle packets received using OpenTTDLib
 *
 * You can use this class freely for your own site.
 *  parts of this file are derived from a file previously maintained by OpenTTD and was therefore freleased under the GNU GPL v2.
 *  basicly you are free to alter it how ever you like, but
 *  you should always make your code available on request.
 *
 * Please refer to docs/ or {@link http://codecubes.org/wiki/OpenTTDLib} for help and examples
 *
 * please report bugs, problems and feature requests at {@link http://codecubes.org}
 *
 * @author	dihedral <dihedral@codecubes.org>
 * @author	TrueLight <truelight@openttd.org>
 * @link	http://codecubes.org
 * @version $Id: OpenTTDLibPacket.php 273 2009-04-01 10:41:19Z dih $
 * @license	http://www.gnu.org/licenses/gpl.html GNU General Public License v2
 *
 * @package OpenTTDLib
 **/
 
/**
 * OpenTTDLibPacket provides {@link OpenTTDLib} with means for reading from and writing to a udp packet
 * @author	dihedral <dihedral@codecubes.org>
 * @author	TrueLight <truelight@openttd.org>
 * @license	http://www.gnu.org/licenses/gpl.html GNU General Public License v2
 *
 * @package OpenTTDLib
 **/
class OpenTTDLibPacket
{
	/**#@+
	 * upd packet enum as found in network_udp.c (0.5) or src/network/core/udp.h (trunk)
	 **/
	const UDP_CLIENT_FIND_SERVER	= 0;
	const UDP_SERVER_RESPONSE		= 1;
	const UDP_CLIENT_DETAIL_INFO	= 2;
	const UDP_SERVER_DETAIL_INFO	= 3;
	const UDP_CLIENT_GET_NEWGRFS	= 9;
	const UDP_SERVER_NEWGRFS		= 10;
	/**#@-*/
	
	/**
	 * size of a packet
	 **/
	const SEND_MTU					= 1460;
   
	/**
	 * @access private
	 * @var string $_rev revision number
	 **/
	private $_rev = '$Rev: 273 $';
	
	/**
	 * @access private
	 * @var string $data the received raw packet
	 **/
	private $_data;
	
	/**
	 * @access private
	 * @var integer size read out of the packet (first 2 bytes)
	 **/
	private $_size;
	
	/**
	 * @access private
	 * @var integer $pos internal position where to start next read/write
	 **/
	private $_pos;
	
	/**
	 * initiate the packet instance
	 * @access	public
	 * @param	string	$data	the packet data
	 * @param	integer	$size	size of the packet
	 * @param	integer $pos	postition next read will start at
	 **/
	public function __construct( $data="", $size=0, $pos=0 )
	{
		/* on an empty packet instance, reserve first 2 bytes for packet size and advance pointer*/
		if( empty( $data ) && $size == 0 && $pos == 0 ) {
			/* = '00' makes sure $this->_data remains a string when appending data */
			$this->_data = '00';
			$this->_pos = 2;
		}
		else {
			$this->_data	= $data;
			$this->_size	= $size;
			$this->_pos		= $pos;
		}
	}
	
	/**
	 * inialize a packet to add data too for later sending
	 * @access public
	 * @param integer $type packet type, one of {@link OPENTTDLIB_PACKET_UPP_CLIENT_FIND_SERVER} or {@link OPENTTDLIB_PACKET_UDP_CLIENT_DETAIL_INFO}
	 * @return OpenTTDLibPacket instance
	 **/
	public function &NetworkSend_Init( $type )
	{
		$packet = new OpenTTDLibPacket();
		
		$packet->Send_uint8( $type );
		
		return $packet;
	}
	
	/**
	 * write packetsize to the packet itself (first 2 bytes)
	 * @return void
	 **/
	public function PrepareToSend()
	{
		$this->_size = $this->_pos;
		$this->_data[0] = chr( $this->_size & 0xFF );
		$this->_data[1] = chr( ( $this->_size >> 8 ) & 0xFF );
		
		/* make sure the size we just wrote to the packet is actually good */
		$this->sizeOK();
	}
	
	/**#@+
	 * append data to the packet for sending
	 * @access	public
	 * @return	void
	 **/
	/**
	 * @param integer $d integer to append to packet
	 **/
	public function Send_uint8( $d )
	{
		$this->_data[$this->_pos++] = chr( $d );
	}
	
	/**
	 * @param integer $d integer to append to packet
	 * @param boolean $switch_endian reverse endianness
	 **/
	public function Send_uint16( $d, $switch_endian=false )
	{
		if( $switch_endian )
		{
			$this->Send_Uint8( ( $d >> 8 ) & 0xFF );
			$this->Send_uint8( $d & 0xFF );
		}
		else
		{
			$this->Send_uint8( $d & 0xFF );
			$this->Send_Uint8( ( $d >> 8 ) & 0xFF );
		}
	}
	
	/**
	 * @param integer $d integer to append to packet
	 * @param boolean $switch_endian reverse endianness
	 **/
	public function Send_uint32( $d, $switch_endian=false )
	{
		if( $switch_endian )
		{
			$this->Send_uint16( $d >> 16, true );
			$this->Send_uint16( $d, true );
		}
		else
		{
			$this->Send_uint16( $d );
			$this->Send_uint16( $d >> 16 );
		}
	}
	
	/**
	 * @param string $s String to append to packet
	 **/
	public function Send_string( $s )
	{
		for( $i = 0; $i < strlen( $s ); $i++ ) {
			$this->_data[$this->_pos++] = $s[$i];
		}
		
		$this->_data[$this->_pos++] = 0;
	}
	/**#@-*/
	
	/**#@+
	 * read the next few byets, process and return a value we can work with
	 * @access	public
	 * @return	integer
	 **/
	public function Recv_uint8()
	{
		return ord( $this->_data[$this->_pos++] );
	}
	
	/**
	 * @param boolean $switch_endian reverse endianness
	 **/
	public function Recv_uint16( $switch_endian=false )
	{
		if( $switch_endian )
		{
			$res = $this->Recv_uint8() << 8;
			$res += $this->Recv_uint8();
		}
		else
		{
			$res = $this->Recv_uint8();
			$res += $this->Recv_uint8() << 8;
		}
		
		return $res;
	}

	/**
	 * @param boolean $switch_endian reverse endianness
	 **/
	public function Recv_uint32( $switch_endian = false )
	{
		if( $switch_endian )
		{
			$res = $this->Recv_uint16( true ) << 16;
			$res += $this->Recv_uint16( true );
		}
		else
		{
			$res = $this->Recv_uint16();
			$res += $this->Recv_uint16() << 16;
		}
		
		return $res;
	}
	/**#@-*/
	
	/**#@+
	 * read the next few bytes, process and return a value we can work with
	 * @access	public
	 * @return	string
	 **/
	/**
	 * 64 bit integer handling using {@link http://www.php.net/manual/en/ref.bc.php BCMath} extentions
	 * 
	 * <br />
	 * <br />please use the BCMath functions for handling 64 bit integers
	 * <br /><b>thanks to eekee, Rubidium and SmatZ</b>
	 *
	 * @internal do not change this method
	 * @link http://www.php.net/manual/en/ref.bc.php
	 **/
	public function Recv_uint64( )
	{
		/* make sure we work with a string later on when using BCMath functions */
		$res = "";
		
		$res_array[1] = $this->Recv_uint8();
		$res_array[8] = $this->Recv_uint8();
		$res_array[16] = $this->Recv_uint8();
		$res_array[24] = $this->Recv_uint8();
		$res_array[32] = $this->Recv_uint8();
		$res_array[40] = $this->Recv_uint8();
		$res_array[48] = $this->Recv_uint8();
		$res_array[56] = $this->Recv_uint8();
		
		foreach ( $res_array as $exp => $base ) {
		        $res = bcadd( $res, bcmul( $base, bcpow( 2, $exp ) ) );
		}
		
		/* check if number is negative - thanks to Rubidum */
		if( $res > bcpow( 2, 63 ) ) {
		        $res = bcsub( $res, bcpow( 2, 64 ) );
		}

		return $res;
	}
	
	public function Recv_string()
	{
		$res = "";
		while(ord($this->_data[$this->_pos]) != 0) {
			$res .= $this->_data[$this->_pos++];
		}
		
		/* advance pointer after finding the 0 */
		$this->_pos++;
		
		return $res;
	}
	
	/**#@-*/
	
	/**
	 * resets the internal position pointer to 0 and readas packet size from packet (first 2 bytes)
	 * @access public
	 * @return void
	 **/
	public function ReadRawPacketSize()
	{
		$this->resetPos();
		$this->_size = $this->Recv_uint16();
	}
	
	/**
	 * compare actual packet size with claimed packet size
	 * @access public
	 * @return bool false if the packet sizes are not the same
	 **/
	public function sizeOK()
	{
		/* $this->_size > self::SEND_MTU should not be possible */
		return ( $this->_size == strlen( $this->_data ) && $this->_size <= self::SEND_MTU ) ? true : false;
	}
	
	/**
	 * get the size of the packet
	 * @access	public
	 * @return	integer
	 **/
	public function getSize()
	{
		return $this->_size;
	}
	
	/**
	 * get the postion for next read
	 * @access	public
	 * @return	integer
	 **/
	public function getPos()
	{
		return $this->_pos;
	}
	
	/**
	 * get the actual packet data
	 * @access	public
	 * @return	string
	 **/
	public function getData()
	{
		return $this->_data;
	}
	
	/**
	 * set the internal pointer back to 0
	 * @access	public
	 * @return	void
	 **/
	public function resetPos()
	{
		$this->_pos = 0;
	}
	
	/**
	 * binary safe storing the packet data
	 * @access public
	 * @return array class variables to serialize
	 **/
	public function __sleep()
	{
		$this->_data = base64_encode( $this->_data );
		return array( '_data', '_size', '_pos', '_rev' );
	}
	
	/**
	 * binary safe restoring the packet data
	 * @access public
	 * @return void
	 **/
	public function __wakeup()
	{
		$this->_data = base64_decode( $this->_data );
	}
}

?>
